;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; The combinator birds of Raymond Smullyan's
;; "To Mock a Mockingbird" (Knopf, 1985)
;; ISBN 0-394-53491-3
;;
;; Note that some birds (Θ, Ω, P, ...) are excluded, either
;; because their definitions are impossible or because the
;; author hasn't gotten around to it. Contributions and
;; corrections are welcome at <evhan@thunktastic.com>.
(module combinator-birds *
  (import scheme)

  ;; Implicit currying
  (define-syntax λ
    (syntax-rules ()
      ((_ (<arg> ...) <body> ...)
       (let ((fn (lambda (<arg> ...) <body> ...))
             (len (length '(<arg> ...))))
         (let loop ((i 0) (lst '()))
           (lambda args
             (let ((i (+ (length args) i))
                   (lst (append lst args)))
               (if (< i len)
                 (loop i lst)
                 (apply fn lst)))))))))

  ;; "To Mock a Mockingbird" pp. 73-74
  (define mockingbird   (λ (x) (x x)))

  ;; "Hopeless Egocentricity" pp. 76-78
  (define kestrel       (λ (x y) x))

  ;; "Identity Birds" pp. 78-80
  (define identity-bird (λ (x) x))

  ;; "Larks" pp. 80-81
  (define lark          (λ (x y) (x (y y))))

  ;; "Is There a Sage Bird?" pp. 89-92
  ;; Unfortunately without a lazy scheme the naive
  ;; definition here doesn't work. We can fake it with
  ;; a proper Y for call-by-value, though obviously
  ;; the various derivations given in the text
  ;; (SLL, BML, etc.) will still not work.
  (define sage-bird
    (λ (x) ((λ (y) (x (λ (x) ((y y) x))))
              (λ (y) (x (λ (x) ((y y) x)))))))

  ;; "Bluebirds" pp. 95-96
  (define bluebird      (λ (x y z) (x (y z))))

  ;; "Some Derivatives of the Bluebird" pp. 96-98
  (define dove          (λ (x y z w) ((x y) (z w))))
  (define blackbird     (λ (x y z w) (x ((y z) w))))
  (define eagle         (λ (x y z w v) ((x y) ((z w) v))))
  (define bunting       (λ (x y z w v) (x (((y z) w) v))))
  (define dickcissel    (λ (x y z w v) (((x y) z) (w v))))
  (define becard        (λ (x y z w) (x (y (z w)))))
  (define dovekies      (λ (x y z w v) ((x (y z)) (w v))))
  (define bald-eagle    (λ (x y z w v u t)
                          ((x ((y z) w)) ((v u) t))))

  ;; "Some Other Birds" pp. 99-101
  (define warbler       (λ (x y) ((x y) y)))
  (define cardinal      (λ (x y z) ((x z) y)))
  (define thrush        (λ (x y) (y x)))

  ;; "Bluebirds and Thrushes" pp. 101-103
  (define robin         (λ (x y z) ((y z) x)))
  (define finch         (λ (x y z) ((z y) x)))
  (define vireo         (λ (x y z) ((z x) y)))

  ;; "Some Relatives" pp. 103-105
  (define cardinal*     (λ (x y z w) (((x y) w) z)))
  (define robin*        (λ (x y z w) (((x z) w) y)))
  (define finch*        (λ (x y z w) (((x w) z) y)))
  (define vireo*        (λ (x y z w) (((x w) z) y)))
  (define cardinal**    (λ (x y z w v) ((((x y) z) w) v)))
  (define robin**       (λ (x y z w v) ((((x y) w) v) z)))
  (define finch**       (λ (x y z w v) ((((x y) v) w) z)))
  (define vireo**       (λ (x y z w v) ((((x y) z) w) v)))

  ;; "Queer Birds" pp. 105-108
  (define queer-bird       (λ (x y z) (y (x z))))
  (define quixotic-bird    (λ (x y z) (x (z y))))
  (define quizzical-bird   (λ (x y z) (y (z x))))
  (define quirky-bird      (λ (x y z) (z (x y))))
  (define quacky-bird      (λ (x y z) (z (y x))))
  (define goldfinch        (λ (x y z w) ((x w) (y z))))

  ;; "More on Mockingbirds" pp. 117-118
  (define double-mockingbird  (λ (x y) ((x y) (x y))))

  ;; "Warblers" pp. 118-120
  (define converse-warbler (λ (x y) ((y x) x)))
  (define warbler*         (λ (x y z) (((x y) z) z)))
  (define warbler**        (λ (x y z w) ((((x y) z) w) w)))
  (define hummingbird      (λ (x y z) (((x y) z) y)))

  ;; "Starlings" pp. 120-121
  (define starling         (λ (x y z) ((x z) (y z))))

  ;; "Starlings in Action" pp. 121-125
  (define phoenix          (λ (x y z w) ((x (y w)) (z w))))
  (define psi-bird         (λ (x y z w) ((x (y z)) (y w))))

  ;; "The Turing Bird" pp. 132-133
  (define turing-bird      (λ (x y) (y ((x x) y))))

  ;; "Owls" p. 133
  (define owl              (λ (x y) (y (x y))))

  ;; "Aristocratic Birds" pp. 179-181
  (define jay              (λ (x y z w) ((x y) ((x w) z))))

  ;; Aliases
  (define oracle-bird sage-bird)
  (define idiot-bird identity-bird)

  ;; A few birds and aliases not necessarily given in the book
  ;; but discussed in David Keenan's http://dkeenan.com/Lambda/
  (define identity-bird*   (λ (x y) (x y)))
  (define kite             (λ (x y) y))
  (define konstant-mocker  (λ (x y) (y y)))
  (define crossed-konstant-mocker (λ (x y) (x x)))
  (define idiot-bird* identity-bird*)
  (define pairing vireo))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; The single-letter names of the various combinators
;; as given in "Who's Who Among The Birds", pp. 244-246
(module combinator-letters *
  (import scheme combinator-birds)
  (define B bluebird)
  (define C cardinal)
  (define D dove)
  (define E eagle)
  (define F finch)
  (define G goldfinch)
  (define H hummingbird)
  (define I identity-bird)
  (define J jay)
  (define K kestrel)
  (define L lark)
  (define M mockingbird)
  ;; P is used in "Russell's Forest" pp. 151-153
  (define O owl)
  ;; N is used in "Curry's Lively Bird Forest" pp. 143-147
  (define Q queer-bird)
  (define R robin)
  (define S starling)
  (define T thrush)
  (define U turing-bird)
  (define V vireo)
  (define W warbler)
  ;; X is used in "The Master Forest" pp. 167-178
  (define Y sage-bird))
